#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2023-2024.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------
import warnings
from string import Formatter
from typing import Any, Sequence, List, Dict, Union, Optional
from dataclasses import dataclass, KW_ONLY, asdict
from json import loads as json_loads

from ibm_watsonx_ai._wrappers import requests
from ibm_watsonx_ai.helpers import DataConnection
from ibm_watsonx_ai.messages.messages import Messages
from ibm_watsonx_ai.wml_client_error import WMLClientError, InvalidMultipleArguments
from ibm_watsonx_ai.utils.autoai.utils import load_file_from_file_system_nonautoai
from ibm_watsonx_ai.foundation_models.utils.enums import ModelTypes
from ibm_watsonx_ai.utils.autoai.enums import DataConnectionTypes
from ibm_watsonx_ai.lifecycle import SpecStates
from ibm_watsonx_ai.helpers.connections.base_data_connection import BaseDataConnection
from ibm_watsonx_ai import APIClient


@dataclass
class PromptTuningParams:
    base_model: dict
    _: KW_ONLY
    accumulate_steps: int = None
    batch_size: int = None
    init_method: str = None
    init_text: str = None
    learning_rate: float = None
    max_input_tokens: int = None
    max_output_tokens: int = None
    num_epochs: int = None
    task_id: str = None
    tuning_type: str = None
    verbalizer: str = None

    def to_dict(self):
        return {key: value for key, value in asdict(self).items() if value is not None}


def _get_foundation_models_spec(url, operation_name, additional_params: dict = None):
    params = {"version": "2023-09-30"}
    if additional_params:
        params.update(additional_params)
    response = requests.get(url,
                            params=params,
                            headers={'X-WML-User-Client': 'PythonClient'})
    if response.status_code == 200:
        return response.json()
    elif response.status_code == 404:
        raise WMLClientError(Messages.get_message(message_id="fm_prompt_tuning_no_foundation_models"))
    else:
        msg = f"{operation_name} failed. Reason: {response.text}"
        raise WMLClientError(msg)


def get_model_specs(url: str, model_id: Optional[str] = None) -> dict:
    """
    Operations to retrieve the list of deployed foundation models specifications.

    :param url: environment url
    :type url: str

    :param model_id: Id of the model, defaults to None (all models specs are returned).
    :type model_id: Optional[str, ModelTypes], optional

    :return: list of deployed foundation model specs
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watsonx_ai.foundation_models import get_model_specs

        # GET ALL MODEL SPECS
        get_model_specs(
            url="https://us-south.ml.cloud.ibm.com"
            )

        # GET MODEL SPECS BY MODEL_ID
        get_model_specs(
            url="https://us-south.ml.cloud.ibm.com",
            model_id="google/flan-ul2"
            )
    """
    try:
        if model_id:
            if isinstance(model_id, ModelTypes):
                model_id = model_id.value
            try:
                return [res for res in _get_foundation_models_spec(f"{url}/ml/v1/foundation_model_specs",
                                                                   "Get available foundation models")['resources']
                        if res['model_id'] == model_id][0]
            except WMLClientError:  # Remove on CPD 5.0 release
                return [res for res in _get_foundation_models_spec(f"{url}/ml/v1-beta/foundation_model_specs",
                                                                   "Get available foundation models")['resources']
                        if res['model_id'] == model_id][0]
        else:
            try:
                return _get_foundation_models_spec(f"{url}/ml/v1/foundation_model_specs",
                                                   "Get available foundation models")
            except WMLClientError:  # Remove on CPD 5.0 release
                return _get_foundation_models_spec(f"{url}/ml/v1-beta/foundation_model_specs",
                                                   "Get available foundation models")
    except WMLClientError as e:
        raise WMLClientError(Messages.get_message(url, message_id="fm_prompt_tuning_no_model_specs"), e)


def get_custom_model_specs(credentials: dict = None,
                           api_client: 'APIClient' = None,
                           model_id: Optional[str] = None,
                           verify=None) -> dict:
    """Get details on available custom models

    :param credentials: credentials to watsonx.ai instance
    :type credentials: dict, optional

    :param api_client: API client to connect to service
    :type api_client: APIClient, optional

    :param model_id: Id of the model, defaults to None (all models specs are returned).
    :type model_id: str, optional

    :param verify: user can pass as verify one of following:

        - the path to a CA_BUNDLE file
        - the path of directory with certificates of trusted CAs
        - `True` - default path to truststore will be taken
        - `False` - no verification will be made
    :type verify: bool or str, optional

    :return: details of supported custom models
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watsonx_ai.foundation_models import get_custom_model_specs

        get_custom_models_spec(api_client=client)
        get_custom_models_spec(credentials=credentials)
        get_custom_models_spec(api_client=client, model_id='mistralai/Mistral-7B-Instruct-v0.2')

    """
    warnings.warn("Model needs to be first stored via client.repository.store_model(model_id, meta_props=metadata)"
                  " and deployed via client.deployments.create(asset_id, metadata) to be used.")

    if credentials is None and api_client is None:
        raise InvalidMultipleArguments(params_names_list=["credentials", "api_client"],
                                       reason="None of the arguments were provided.")
    elif credentials:
        client = APIClient(credentials, verify=verify)
    else:
        client = api_client

    url = client.wml_credentials['url']
    token = client.wml_token

    if token is None:
        token = BaseDataConnection._get_token_from_environment()

    response = requests.get(f"{url}/ml/v4/custom_foundation_models",
                            params=client._params(skip_for_create=True),
                            headers={'Authorization': f"Bearer {token}"})
    if response.status_code == 200:
        if model_id:
            resources = [res for res in response.json()['resources'] if res['model_id'] == model_id]
            return resources[0] if resources else {}
        else:
            return response.json()
    elif response.status_code == 404:
        raise WMLClientError(Messages.get_message(url, message_id="custom_models_no_model_specs"), url)
    else:
        msg = f"Getting failed. Reason: {response.text}"
        raise WMLClientError(msg)


def get_model_lifecycle(url: str, model_id: str) -> Union[list, None]:
    """
    Operation to retrieve the list of model lifecycle data.

    :param url: environment url
    :type url: str

    :param model_id: the type of model to use
    :type model_id: str

    :return: list of deployed foundation model lifecycle data
    :rtype: list

    **Example**

    .. code-block:: python

        from ibm_watsonx_ai.foundation_models import get_model_lifecycle

        get_model_lifecycle(
            url="https://us-south.ml.cloud.ibm.com",
            model_id="ibm/granite-13b-instruct-v2"
            )
    """
    model_specs = get_model_specs(url=url)
    model_spec = next((model_metadata for model_metadata in model_specs.get('resources', [])
                       if model_metadata.get('model_id') == model_id), None)
    return model_spec.get('lifecycle') if model_spec is not None else None


def _check_model_state(url, model_id):
    default_warning_template = (
        "Model '{model_id}' is in {state} state from {start_date} until {withdrawn_start_date}. "
        "IDs of alternative models: {alternative_model_ids}. "
        "Further details: https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-model-lifecycle.html?context=wx&audience=wdp")

    lifecycle = get_model_lifecycle(url, model_id)

    modes_list = [ids.get("id") for ids in (lifecycle or [])]

    if lifecycle and SpecStates.DEPRECATED.value in modes_list and SpecStates.CONSTRICTED.value in modes_list:
        deprecated_and_constricted_warning_template = (
            "Model '{model_id}' is in deprecated state from {deprecated_start_date} until the model status is changed "
            "to constricted {constricted_start_date} state, which will last until {withdrawn_start_date}. "
            "IDs of alternative models: {alternative_model_ids}. "
            "Further details: https://dataplatform.cloud.ibm.com/docs/content/wsj/analyze-data/fm-model-lifecycle.html?context=wx&audience=wdp")
        warnings.warn(deprecated_and_constricted_warning_template.format(
            model_id=model_id,
            deprecated_start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.DEPRECATED.value), None),
            constricted_start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.CONSTRICTED.value), None),
            withdrawn_start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.WITHDRAWN.value), None),
            alternative_model_ids=next((', '.join(el.get('alternative_model_ids', ["None"])) for el in lifecycle
                                        if el.get('id') == SpecStates.CONSTRICTED.value), None)
        ), category=LifecycleWarning)

    elif lifecycle and SpecStates.DEPRECATED.value in modes_list:
        warnings.warn(default_warning_template.format(
            model_id=model_id,
            state=SpecStates.DEPRECATED.value,
            start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.DEPRECATED.value), None),
            withdrawn_start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.WITHDRAWN.value), None),
            alternative_model_ids=next((', '.join(el.get('alternative_model_ids', ["None"])) for el in lifecycle
                                        if el.get('id') == SpecStates.DEPRECATED.value), None)
        ), category=LifecycleWarning)

    elif lifecycle and SpecStates.CONSTRICTED.value in modes_list:
        warnings.warn(default_warning_template.format(
            model_id=model_id,
            state=SpecStates.CONSTRICTED.value,
            start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.CONSTRICTED.value), None),
            withdrawn_start_date=next((el.get('start_date') for el in lifecycle if el.get('id') == SpecStates.WITHDRAWN.value), None),
            alternative_model_ids=next((', '.join(el.get('alternative_model_ids', ["None"])) for el in lifecycle
                                        if el.get('id') == SpecStates.CONSTRICTED.value), None)
        ), category=LifecycleWarning)


def get_model_specs_with_prompt_tuning_support(url: str) -> dict:
    """
    Operations to query the details of the deployed foundation models with prompt tuning support.

    :param url: environment url
    :type url: str

    :return: list of deployed foundation model specs with prompt tuning support
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watsonx_ai.foundation_models import get_model_specs_with_prompt_tuning_support

        get_model_specs_with_prompt_tuning_support(
            url="https://us-south.ml.cloud.ibm.com"
            )
    """
    try:
        try:
            return _get_foundation_models_spec(url=f"{url}/ml/v1/foundation_model_specs",
                                               operation_name="Get available foundation models",
                                               additional_params={"filters": "function_prompt_tune_trainable"})
        except WMLClientError:  # Remove on CPD 5.0 release
            return _get_foundation_models_spec(url=f"{url}/ml/v1-beta/foundation_model_specs",
                                               operation_name="Get available foundation models",
                                               additional_params={"filters": "function_prompt_tune_trainable"})
    except WMLClientError as e:
        raise WMLClientError(Messages.get_message(url, message_id="fm_prompt_tuning_no_model_specs"), e)


def get_supported_tasks(url: str) -> dict:
    """
    Operation to retrieve the list of tasks that are supported by the foundation models.

    :param url: environment url
    :type url: str

    :return: list of tasks that are supported by the foundation models
    :rtype: dict

    **Example**

    .. code-block:: python

        from ibm_watsonx_ai.foundation_models import get_supported_tasks

        get_supported_tasks(
            url="https://us-south.ml.cloud.ibm.com"
            )
    """
    try:
        try:
            return _get_foundation_models_spec(f"{url}/ml/v1/foundation_model_tasks",
                                               "Get tasks that are supported by the foundation models.")
        except WMLClientError:  # Remove on CPD 5.0 release
            return _get_foundation_models_spec(f"{url}/ml/v1-beta/foundation_model_tasks",
                                               "Get tasks that are supported by the foundation models.")
    except WMLClientError as e:
        raise WMLClientError(Messages.get_message(url, message_id="fm_prompt_tuning_no_supported_tasks"), e)


def get_all_supported_tasks_dict(url="https://us-south.ml.cloud.ibm.com") -> dict:
    tasks_dict = dict()
    for task_spec in get_supported_tasks(url).get('resources', []):
        tasks_dict[task_spec['label'].replace("-", "_").replace(" ", "_").upper()] = task_spec['task_id']
    return tasks_dict


def load_request_json(run_id, wml_client, run_params = None) -> dict:
    if run_params is None:
        run_params = wml_client.training.get_details(run_id)

    model_request_path = run_params['entity'].get('results_reference', {}).get('location', {}).get('model_request_path')

    if model_request_path is None:
        raise WMLClientError("Missing model_request_path in run_params. Verify if the training run has been completed.")

    if wml_client.CLOUD_PLATFORM_SPACES:
        results_reference = DataConnection._from_dict(run_params['entity']['results_reference'])

        if run_params['entity']['results_reference']['type'] == DataConnectionTypes.CA:
            results_reference.location.file_name = model_request_path
        else:
            results_reference.location.path = model_request_path

        results_reference.set_client(wml_client)
        request_json_bytes = results_reference.read(raw=True, binary=True)

        # download from cos

    elif wml_client.CPD_version >= 4.8:
        asset_parts = model_request_path.split('/')
        model_request_asset_url = '/'.join(asset_parts[asset_parts.index('assets') + 1:])
        request_json_bytes = load_file_from_file_system_nonautoai(api_client=wml_client,
                                                                  file_path=model_request_asset_url).read()
    else:
        raise WMLClientError("Unsupported environment for this action")

    return json_loads(request_json_bytes.decode())


def is_training_prompt_tuning(training_id, wml_client):
    """Returns True if training_id is connected to prompt tuning"""
    if training_id is None:
        return False
    run_params = wml_client.training.get_details(training_uid=training_id)
    return bool(run_params['entity'].get('prompt_tuning'))


class TemplateFormatter(Formatter):
    def check_unused_args(self, 
                          used_args: List[(int | str)], 
                          args: Sequence, 
                          kwargs: Dict[str, Any]) -> None:
        """Check for unused args."""
        extra_args = set(kwargs).difference(used_args)
        if extra_args:
            raise KeyError(extra_args)

class HAPDetectionWarning(UserWarning): ...
class PIIDetectionWarning(UserWarning): ...
class LifecycleWarning(UserWarning): ...